'''
Created on Jan 28, 2016

@author: jason
'''
import numpy as np
from pyearth._qr import UpdatingQT


def test_updating_qt():
    np.random.seed(0)
    m = 10
    n = 3

    X = np.random.normal(size=(n, m)).T
    u = UpdatingQT.alloc(m, n, 1e-14)

#     u2 = UpdatingQT(m, n)
    Q = np.linalg.qr(X, mode='reduced')[0]
    u.update(X[:, 0])

#     u2.update(X[:,0])
    u.update(X[:, 1])
#     u2.update(X[:,1])
    u.update(X[:, 2])
#     u2.update(X[:,2])

#     assert np.max(np.abs(np.abs(u2.Q_t) - np.abs(Q.T))) < .0000000000001
    assert np.max(np.abs(np.abs(u.Q_t) - np.abs(Q.T))) < .0000000000001

    X2 = X.copy()
    X2[:, 2] = np.random.normal(size=m)

    u.downdate()
    u.update(X2[:, 2])

    Q2 = np.linalg.qr(X2, mode='reduced')[0]
    assert np.max(np.abs(np.abs(u.Q_t) - np.abs(Q2.T))) < .0000000000001


def test_updating_qr_with_linear_dependence():
    np.random.seed(0)
    m = 10
    n = 5
    assert n >= 3
    X = np.random.normal(size=(n, m)).T
    X[:, 2] = X[:, 0] + 3 * X[:, 1]

    Q = np.linalg.qr(X, mode='reduced')[0]
    u = UpdatingQT.alloc(m, n, 1e-14)
    u2 = UpdatingQT.alloc(m, n, 1e-14)

    u.update(X[:, 0])
    u2.update(X[:, 0])
#     u2.update(X[:,0])
    u.update(X[:, 1])
    u2.update(X[:, 1])
#     u2.update(X[:,1])
    u.update(X[:, 2])

    assert np.max(
        np.abs(np.abs(u.Q_t[:2, :]) - np.abs(Q[:, :2].T))) < .0000000000001
    assert np.max(np.abs(u.Q_t[2, :])) == 0.

    # Make sure you can downdate a dependent column safely
    u.downdate()
    u.update(X[:, 2])
    assert np.max(
        np.abs(np.abs(u.Q_t[:2, :]) - np.abs(Q[:, :2].T))) < .0000000000001
    assert np.max(np.abs(u.Q_t[2, :])) == 0.

    for j in range(3, n):
        u.update(X[:, j])
        u2.update(X[:, j])

    # Q_t is orthonormal except for its zero column
    Q_nonzero = np.concatenate([u.Q_t[:2, :].T, u.Q_t[3:, :].T], axis=1)
    np.testing.assert_array_almost_equal(
        np.dot(Q_nonzero.T, Q_nonzero), np.eye(n - 1))

    # Q_t.T is in the column space of X
    b = np.linalg.lstsq(X, u.Q_t.T)[0]
    Q_hat = np.dot(X, b)
    np.testing.assert_array_almost_equal(Q_hat, u.Q_t.T)

    # X is in the column space of Q_t.T
    a = np.linalg.lstsq(u.Q_t.T, X)[0]
    X_hat = np.dot(u.Q_t.T, a)
    np.testing.assert_array_almost_equal(X_hat, X)

    # u and u2 should have the same householder
    np.testing.assert_array_almost_equal(
        u.householder.V[:, :u.householder.k],
        u2.householder.V[:, :u2.householder.k])
    np.testing.assert_array_almost_equal(
            u.householder.T[:u.householder.k, :u.householder.k],
            u2.householder.T[:u2.householder.k, :u2.householder.k])

    # u should have one more column than u2
    assert u.k == u2.k + 1



from pyearth._record import PruningPassRecord, PruningPassIteration
from pyearth._util import gcv
from ..testing_utils import assert_list_almost_equal


num_samples = 1000
num_variables = 10
penalty = 3.0
sst = 100.0
record = PruningPassRecord(num_samples, num_variables,
                           penalty, sst, 7, 18.0)
record.append(PruningPassIteration(2, 6, 25.0))
record.append(PruningPassIteration(1, 5, 34.0))
record.append(PruningPassIteration(3, 4, 87.0))
mses = [18.0, 25.0, 34.0, 87.0]
sizes = [7, 6, 5, 4]


def test_statistics():
    mses = [record.mse(i) for i in range(len(record))]
    mses_ = [mses[i] for i in range(len(record))]
    gcvs = [record.gcv(i) for i in range(len(record))]
    gcvs_ = [gcv(mses[i], sizes[i], num_samples, penalty)
             for i in range(len(record))]
    rsqs = [record.rsq(i) for i in range(len(record))]
    rsqs_ = [1 - (mses[i] / sst)
             for i in range(len(record))]
    grsqs = [record.grsq(i) for i in range(len(record))]
    grsqs_ = [1 - (record.gcv(i) / gcv(sst, 1, num_samples, penalty))
              for i in range(len(record))]
    assert_list_almost_equal(mses, mses_)
    assert_list_almost_equal(gcvs, gcvs_)
    assert_list_almost_equal(rsqs, rsqs_)
    assert_list_almost_equal(grsqs, grsqs_)



cimport numpy as cnp
from _types cimport FLOAT_t, INT_t, INDEX_t, BOOL_t
from _basis cimport Basis
from _record cimport PruningPassRecord

cdef class PruningPasser:
    cdef cnp.ndarray X
    cdef cnp.ndarray missing
    cdef cnp.ndarray B
    cdef cnp.ndarray y
    cdef cnp.ndarray sample_weight
    cdef int verbose
    cdef cnp.ndarray output_weight
    cdef public dict feature_importance
    
    cdef INDEX_t m
    cdef INDEX_t n
    cdef Basis basis
    cdef FLOAT_t penalty
    cdef FLOAT_t sst
    cdef PruningPassRecord record

    cpdef run(PruningPasser self)

    cpdef PruningPassRecord trace(PruningPasser self)



import pickle
import numpy

from nose.tools import assert_equal

from .base import BaseContainer
from pyearth._types import BOOL
from pyearth._basis import SmoothedHingeBasisFunction, ConstantBasisFunction


class Container(BaseContainer):

    def __init__(self):
        super(Container, self).__init__()
        self.parent = ConstantBasisFunction()
        self.bf1 = SmoothedHingeBasisFunction(self.parent,
                                              1.0, 0.0, 3.0, 10, 1,
                                              False)
        self.bf2 = SmoothedHingeBasisFunction(self.parent,
                                              1.0, 0.0, 3.0, 10, 1,
                                              True)


def test_getters():
    cnt = Container()
    assert not cnt.bf1.get_reverse()
    assert cnt.bf2.get_reverse()
    assert cnt.bf1.get_knot() == 1.0
    assert cnt.bf1.get_variable() == 1
    assert cnt.bf1.get_knot_idx() == 10
    assert cnt.bf1.get_parent() == cnt.parent
    assert cnt.bf1.get_knot_minus() == 0.0
    assert cnt.bf1.get_knot_plus() == 3.0


def test_pickle_compatibility():
    cnt = Container()
    bf_copy = pickle.loads(pickle.dumps(cnt.bf1))
    assert_equal(cnt.bf1, bf_copy)


def test_smoothed_version():
    cnt = Container()
    translation = {cnt.parent: cnt.parent._smoothed_version(None, {}, {})}
    smoothed = cnt.bf1._smoothed_version(cnt.parent, {}, translation)
    assert_equal(cnt.bf1, smoothed)


def test_degree():
    cnt = Container()
    assert_equal(cnt.bf1.degree(), 1)
    assert_equal(cnt.bf2.degree(), 1)


def test_p_r():
    cnt = Container()
    pplus = (2 * 3.0 + 0.0 - 3 * 1.0) / ((3.0 - 0.0)**2)
    rplus = (2 * 1.0 - 3.0 - 0.0) / ((3.0 - 0.0)**3)
    pminus = (3 * 1.0 - 2 * 0.0 - 3.0) / ((0.0 - 3.0)**2)
    rminus = (0.0 + 3.0 - 2 * 1.0) / ((0.0 - 3.0)**3)
    assert_equal(cnt.bf1.get_p(), pplus)
    assert_equal(cnt.bf1.get_r(), rplus)
    assert_equal(cnt.bf2.get_p(), pminus)
    assert_equal(cnt.bf2.get_r(), rminus)


def test_apply():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    B = numpy.ones(shape=(m, 10))
    cnt.bf1.apply(cnt.X, missing, B[:, 0])
    cnt.bf2.apply(cnt.X, missing, B[:, 1])
    pplus = (2 * 3.0 + 0.0 - 3 * 1.0) / ((3.0 - 0.0)**2)
    rplus = (2 * 1.0 - 3.0 - 0.0) / ((3.0 - 0.0)**3)
    pminus = (3 * 1.0 - 2 * 0.0 - 3.0) / ((0.0 - 3.0)**2)
    rminus = (0.0 + 3.0 - 2 * 1.0) / ((0.0 - 3.0)**3)
    c1 = numpy.ones(m)
    c1[cnt.X[:, 1] <= 0.0] = 0.0
    c1[(cnt.X[:, 1] > 0.0) & (cnt.X[:, 1] < 3.0)] = (
        pplus * ((cnt.X[(cnt.X[:, 1] > 0.0) & (
            cnt.X[:, 1] < 3.0), 1] - 0.0)**2) +
        rplus * ((cnt.X[(cnt.X[:, 1] > 0.0) & (
            cnt.X[:, 1] < 3.0), 1] - 0.0)**3))
    c1[cnt.X[:, 1] >= 3.0] = cnt.X[cnt.X[:, 1] >= 3.0, 1] - 1.0
    c2 = numpy.ones(m)
    c2[cnt.X[:, 1] >= 3.0] = 0.0
    c2.flat[cnt.X[:, 1] <= 0.0] = -1 * (cnt.X[cnt.X[:, 1] <= 0.0] - 1.0)
    c2[(cnt.X[:, 1] > 0.0) & (cnt.X[:, 1] < 3.0)] = (
        pminus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                         (cnt.X[:, 1] < 3.0), 1] - 3.0)**2) +
        rminus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                         (cnt.X[:, 1] < 3.0), 1] - 3.0)**3)
    )
    numpy.testing.assert_almost_equal(B[:, 0], c1)
    numpy.testing.assert_almost_equal(B[:, 1], c2)


def test_apply_deriv():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    pplus = (2 * 3.0 + 0.0 - 3 * 1.0) / ((3.0 - 0.0)**2)
    rplus = (2 * 1.0 - 3.0 - 0.0) / ((3.0 - 0.0)**3)
    pminus = (3 * 1.0 - 2 * 0.0 - 3.0) / ((0.0 - 3.0)**2)
    rminus = (0.0 + 3.0 - 2 * 1.0) / ((0.0 - 3.0)**3)
    c1 = numpy.ones(m)
    c1[cnt.X[:, 1] <= 0.0] = 0.0
    c1[(cnt.X[:, 1] > 0.0) & (cnt.X[:, 1] < 3.0)] = (
        pplus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                        (cnt.X[:, 1] < 3.0), 1] - 0.0)**2) +
        rplus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                        (cnt.X[:, 1] < 3.0), 1] - 0.0)**3))
    c1[cnt.X[:, 1] >= 3.0] = cnt.X[cnt.X[:, 1] >= 3.0, 1] - 1.0
    c2 = numpy.ones(m)
    c2[cnt.X[:, 1] >= 3.0] = 0.0
    c2.flat[cnt.X[:, 1] <= 0.0] = -1 * (cnt.X[cnt.X[:, 1] <= 0.0] - 1.0)
    c2[(cnt.X[:, 1] > 0.0) & (cnt.X[:, 1] < 3.0)] = (
        pminus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                         (cnt.X[:, 1] < 3.0), 1] - 3.0)**2) +
        rminus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                         (cnt.X[:, 1] < 3.0), 1] - 3.0)**3)
    )
    b1 = numpy.empty(shape=m)
    j1 = numpy.empty(shape=m)
    b2 = numpy.empty(shape=m)
    j2 = numpy.empty(shape=m)
    cp1 = numpy.ones(m)
    cp1[cnt.X[:, 1] <= 0.0] = 0.0
    cp1[(cnt.X[:, 1] > 0.0) & (cnt.X[:, 1] < 3.0)] = (
        2.0 * pplus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                              (cnt.X[:, 1] < 3.0), 1] - 0.0)) +
        3.0 * rplus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                              (cnt.X[:, 1] < 3.0), 1] - 0.0)**2)
    )
    cp1[cnt.X[:, 1] >= 3.0] = 1.0
    cp2 = numpy.ones(m)
    cp2[cnt.X[:, 1] >= 3.0] = 0.0
    cp2[cnt.X[:, 1] <= 0.0] = -1.0
    cp2[(cnt.X[:, 1] > 0.0) & (cnt.X[:, 1] < 3.0)] = (
        2.0 * pminus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                               (cnt.X[:, 1] < 3.0), 1] - 3.0)) +
        3.0 * rminus * ((cnt.X[(cnt.X[:, 1] > 0.0) &
                               (cnt.X[:, 1] < 3.0), 1] - 3.0)**2))
    cnt.bf1.apply_deriv(cnt.X, missing, b1, j1, 1)
    cnt.bf2.apply_deriv(cnt.X, missing, b2, j2, 1)
    numpy.testing.assert_almost_equal(b1, c1)
    numpy.testing.assert_almost_equal(b2, c2)
    numpy.testing.assert_almost_equal(j1, cp1)
    numpy.testing.assert_almost_equal(j2, cp2)



from ._forward import ForwardPasser
from ._pruning import PruningPasser, FEAT_IMP_CRITERIA
from ._util import ascii_table, apply_weights_2d, gcv
from ._types import BOOL
from sklearn.base import RegressorMixin, BaseEstimator, TransformerMixin
from sklearn.utils.validation import (assert_all_finite, check_is_fitted,
                                      check_X_y)
import numpy as np
from scipy import sparse
from ._version import get_versions
__version__ = get_versions()['version']

class Earth(BaseEstimator, RegressorMixin, TransformerMixin):

    """
    Multivariate Adaptive Regression Splines

    A flexible regression method that automatically searches for interactions
    and non-linear relationships.  Earth models can be thought of as
    linear models in a higher dimensional basis space
    (specifically, a multivariate truncated power spline basis).
    Each term in an Earth model is a product of so called "hinge functions".
    A hinge function is a function that's equal to its argument where that
    argument is greater than zero and is zero everywhere else.

    The multivariate adaptive regression splines algorithm has two stages.
    First, the forward pass searches for terms in the truncated power spline
    basis that locally minimize the squared error loss of the training set.
    Next, a pruning pass selects a subset of those terms that produces
    a locally minimal generalized cross-validation (GCV) score.  The GCV score
    is not actually based on cross-validation, but rather is meant to
    approximate a true cross-validation score by penalizing model complexity.
    The final result is a set of terms that is nonlinear in the original
    feature space, may include interactions, and is likely to generalize well.

    The Earth class supports dense input only.  Data structures from the
    pandas and patsy modules are supported, but are copied into numpy arrays
    for computation.  No copy is made if the inputs are numpy float64 arrays.
    Earth objects can be serialized using the pickle module and copied
    using the copy module.


    Parameters
    ----------
    max_terms : int, optional (default=min(2 * n + m // 10, 400)), 
                               where n is the number of features and m is the number 
                               of rows)
        The maximum number of terms generated by the forward pass.  All memory is
        allocated at the beginning of the forward pass, so setting max_terms to 
        a very high number on a system with insufficient memory may cause a 
        MemoryError at the start of the forward pass.


    max_degree : int, optional (default=1)
        The maximum degree of terms generated by the forward pass.


    allow_missing : boolean, optional (default=False)
        If True, use missing data method described in [3].
        Use missing argument to determine missingness or,if X is a pandas
        DataFrame, infer missingness from X.


    penalty : float, optional (default=3.0)
        A smoothing parameter used to calculate GCV and GRSQ.
        Used during the pruning pass and to determine whether to add a hinge
        or linear basis function during the forward pass.
        See the d parameter in equation 32, Friedman, 1991.


    endspan_alpha : float, optional, probability between 0 and 1 (default=0.05)
        A parameter controlling the calculation of the endspan
        parameter (below).  The endspan parameter is calculated as
        round(3 - log2(endspan_alpha/n)), where n is the number of features.
        The endspan_alpha parameter represents the probability of a run of
        positive or negative error values on either end of the data vector
        of any feature in the data set.  See equation 45, Friedman, 1991.


    endspan : int, optional (default=-1)
        The number of extreme data values of each feature not eligible
        as knot locations. If endspan is set to -1 (default) then the
        endspan parameter is calculated based on endspan_alpah (above).
        If endspan is set to a positive integer then endspan_alpha is ignored.


    minspan_alpha : float, optional, probability between 0 and 1 (default=0.05)
        A parameter controlling the calculation of the minspan
        parameter (below).  The minspan parameter is calculated as

            (int) -log2(-(1.0/(n*count))*log(1.0-minspan_alpha)) / 2.5

        where n is the number of features and count is the number of points at
        which the parent term is non-zero.  The minspan_alpha parameter
        represents the probability of a run of positive or negative error
        values between adjacent knots separated by minspan intervening
        data points. See equation 43, Friedman, 1991.


    minspan : int, optional (default=-1)
        The minimal number of data points between knots.  If minspan is set
        to -1 (default) then the minspan parameter is calculated based on
        minspan_alpha (above).  If minspan is set to a positive integer then
        minspan_alpha is ignored.


    thresh : float, optional (default=0.001)
        Parameter used when evaluating stopping conditions for the forward
        pass. If either RSQ > 1 - thresh or if RSQ increases by less than
        thresh for a forward pass iteration then the forward pass is
        terminated.


    zero_tol : float, optional (default=1e-12)
        Used when determining whether a floating point number is zero during
        the  forward pass.  This is important in determining linear dependence
        and in the fast update procedure.  There should normally be no reason
        to change  zero_tol from its default. However, if nans are showing up
        during the forward pass or the forward pass seems to be terminating
        unexpectedly, consider adjusting zero_tol.

    min_search_points : int, optional (default=100)
        Used to calculate check_every (below).  The minimum samples necessary
        for check_every to be greater than 1.  The check_every parameter
        is calculated as

             (int) m / min_search_points

        if m > min_search_points, where m is the number of samples in the
        training set.  If m <= min_search_points then check_every is set to 1.


    check_every : int, optional (default=-1)
        If check_every > 0, only one of every check_every sorted data points
        is considered as a candidate knot.  If check_every is set to -1 then
        the check_every parameter is calculated based on
        min_search_points (above).


    allow_linear : bool, optional (default=True)
        If True, the forward pass will check the GCV of each new pair of terms
        and, if it's not an improvement on a single term with no knot (called a
        linear term, although it may actually be a product of a linear term
        with some other parent term), then only that single, knotless term will
        be used. If False, that behavior is disabled and all terms will have
        knots except those with variables specified by the linvars argument
        (see the fit method).

    use_fast : bool, optional (default=False)
        if True, use the approximation procedure defined in [2] to speed up the
        forward pass. The procedure uses two hyper-parameters : fast_K
        and fast_h. Check below for more details.

    fast_K : int, optional (default=5)
        Only used if use_fast is True. As defined in [2], section 3.0, it
        defines the maximum number of basis functions to look at when
        we search for a parent, that is we look at only the fast_K top
        terms ranked by the mean squared error of the model the last time
        the term was chosen as a parent. The smaller fast_K is, the more
        gains in speed we get but the more approximate is the result.
        If fast_K is the maximum number of terms and fast_h is 1,
        the behavior is the same as in the normal case
        (when use_fast is False).

    fast_h : int, optional (default=1)
        Only used if use_fast is True. As defined in [2], section 4.0, it
        determines the number of iterations before repassing through all
        the variables when searching for the variable to use for a
        given parent term. Before reaching fast_h number of iterations
        only the last chosen variable for the parent term is used. The
        bigger fast_h is, the more speed gains we get, but the result
        is more approximate.

    smooth : bool, optional (default=False)
        If True, the model will be smoothed such that it has continuous first
        derivatives.
        For details, see section 3.7, Friedman, 1991.

    enable_pruning : bool, optional(default=True)
        If False, the pruning pass will be skipped.

    feature_importance_type: string or list of strings, optional (default=None)
        Specify which kind of feature importance criteria to compute.
        Currently three criteria are supported : 'gcv', 'rss' and 'nb_subsets'.
        By default (when it is None), no feature importance is computed.
        Feature importance is a measure of the effect of the features
        on the outputs. For each feature, the values go from
        0 to 1 and sum up to 1. A high value means the feature have in average
        (over the population) a large effect on the outputs.
        See [4], section 12.3 for more information about the criteria.

    verbose : int, optional(default=0)
        If verbose >= 1, print out progress information during fitting.  If
        verbose >= 2, also print out information on numerical difficulties
        if encountered during fitting. If verbose >= 3, print even more
        information that is probably only useful to the developers of py-earth.

    Attributes
    ----------
    `coef_` : array, shape = [pruned basis length, number of outputs]
        The weights of the model terms that have not been pruned.


    `basis_` : _basis.Basis
        An object representing model terms.  Each term is a product of
        constant, linear, and hinge functions of the input features.


    `mse_` : float
        The mean squared error of the model after the final linear fit.
        If sample_weight and/or output_weight are given, this score is
        weighted appropriately.


    `rsq_` : float
        The generalized r^2 of the model after the final linear fit.
        If sample_weight and/or output_weight are given, this score is
        weighted appropriately.

    `gcv_` : float
        The generalized cross validation (GCV) score of the model after the
        final linear fit. If sample_weight and/or output_weight are
        given, this score is weighted appropriately.

    `grsq_` : float
        An r^2 like score based on the GCV. If sample_weight and/or
        output_weight are given, this score is
        weighted appropriately.

    `forward_pass_record_` : _record.ForwardPassRecord
        An object containing information about the forward pass, such as
        training loss function values after each iteration and the final
        stopping condition.


    `pruning_pass_record_` : _record.PruningPassRecord
        An object containing information about the pruning pass, such as
        training loss function values after each iteration and the
        selected optimal iteration.


    `xlabels_` : list
        List of column names for training predictors.
        Defaults to ['x0','x1',....] if column names are not provided.


    `allow_missing_` : list
        List of booleans indicating whether each variable is allowed to
        be missing.  Set during training.  A variable may be missing
        only if fitting included missing data for that variable.

    `feature_importances_`: array of shape [m] or dict
        m is the number of features.
        if one feature importance type is specified, it is an
        array of shape m. If several feature importance types are
        specified, then it is dict where each key is a feature importance type
        name and its corresponding value is an array of shape m.
    
    `_version`: string
        The version of py-earth in which the Earth object was originally 
        created.  This information may be useful when dealing with 
        serialized Earth objects.


    References
    ----------

    .. [1] Friedman, Jerome. Multivariate Adaptive Regression Splines.
           Annals of Statistics. Volume 19, Number 1 (1991), 1-67.

    .. [2] Fast MARS, Jerome H.Friedman, Technical Report No.110, May 1993.

    .. [3] Estimating Functions of Mixed Ordinal and Categorical Variables
           Using Adaptive Splines, Jerome H.Friedman, Technical Report
           No.108, June 1991.

    .. [4] http://www.milbo.org/doc/earth-notes.pdf

    """

    forward_pass_arg_names = set([
        'max_terms', 'max_degree', 'allow_missing', 'penalty',
        'endspan_alpha', 'endspan',
        'minspan_alpha', 'minspan',
        'thresh', 'zero_tol', 'min_search_points',
        'check_every', 'allow_linear',
        'use_fast', 'fast_K', 'fast_h',
        'feature_importance_type',
        'verbose'
    ])
    pruning_pass_arg_names = set([
        'penalty',
        'feature_importance_type',
        'verbose'
    ])

    def __init__(self, max_terms=None, max_degree=None, allow_missing=False,
                 penalty=None, endspan_alpha=None, endspan=None,
                 minspan_alpha=None, minspan=None,
                 thresh=None, zero_tol=None, min_search_points=None,
                 check_every=None,
                 allow_linear=None, use_fast=None, fast_K=None,
                 fast_h=None, smooth=None, enable_pruning=True,
                 feature_importance_type=None, verbose=0):

        self.max_terms = max_terms
        self.max_degree = max_degree
        self.allow_missing = allow_missing
        self.penalty = penalty
        self.endspan_alpha = endspan_alpha
        self.endspan = endspan
        self.minspan_alpha = minspan_alpha
        self.minspan = minspan
        self.thresh = thresh
        self.zero_tol = zero_tol
        self.min_search_points = min_search_points
        self.check_every = check_every
        self.allow_linear = allow_linear
        self.use_fast = use_fast
        self.fast_K = fast_K
        self.fast_h = fast_h
        self.smooth = smooth
        self.enable_pruning = enable_pruning
        self.feature_importance_type = feature_importance_type
        self.verbose = verbose
        self._version = __version__

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        keys = set(self.__dict__.keys()) | set(other.__dict__.keys())
        for k in keys:
            try:
                v_self = self.__dict__[k]
                v_other = other.__dict__[k]
            except KeyError:
                return False
            try:
                if v_self != v_other:
                    return False
            except ValueError:  # Case of numpy arrays
                if np.any(v_self != v_other):
                    return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def _pull_forward_args(self, **kwargs):
        '''
        Pull named arguments relevant to the forward pass.
        '''
        result = {}
        for name in self.forward_pass_arg_names:
            if name in kwargs and kwargs[name] is not None:
                result[name] = kwargs[name]
        return result

    def _pull_pruning_args(self, **kwargs):
        '''
        Pull named arguments relevant to the pruning pass.
        '''
        result = {}
        for name in self.pruning_pass_arg_names:
            if name in kwargs and kwargs[name] is not None:
                result[name] = kwargs[name]
        return result

    def _scrape_labels(self, X):
        '''
        Try to get labels from input data (for example, if X is a
        pandas DataFrame).  Return None if no labels can be extracted.
        '''
        try:
            labels = list(X.columns)
        except AttributeError:
            try:
                labels = list(X.design_info.column_names)
            except AttributeError:
                try:
                    labels = list(X.dtype.names)
                except TypeError:
                    try:
                        labels = ['x%d' % i for i in range(X.shape[1])]
                    except IndexError:
                        labels = ['x%d' % i for i in range(1)]
                # handle case where X is not np.array (e.g list)
                except AttributeError:
                    X = np.array(X)
                    labels = ['x%d' % i for i in range(X.shape[1])]
        return labels

    def _scrub_x(self, X, missing, **kwargs):
        '''
        Sanitize input predictors and extract column names if appropriate.
        '''
        # Check for sparseness
        if sparse.issparse(X):
            raise TypeError('A sparse matrix was passed, but dense data '
                            'is required. Use X.toarray() to convert to '
                            'dense.')
        X = np.asarray(X, dtype=np.float64, order='F')
        
        # Figure out missingness
        missing_is_nan = False
        if missing is None:
            # Infer missingness
            missing = np.isnan(X)
            missing_is_nan = True
            
        if not self.allow_missing:
            try:
                assert_all_finite(X)
            except ValueError:
                raise ValueError(
                    "Input contains NaN, infinity or a value that's too large."
                    "Did you mean to set allow_missing=True?")
        if X.ndim == 1:
            X = X[:, np.newaxis]

        # Ensure correct number of columns
        if hasattr(self, 'basis_') and self.basis_ is not None:
            if X.shape[1] != self.basis_.num_variables:
                raise ValueError('Wrong number of columns in X. Reshape your data.')
        
        # Zero-out any missing spots in X
        if np.any(missing):
            if not self.allow_missing:
                raise ValueError('Missing data requires allow_missing=True.')
            if missing_is_nan or np.any(np.isnan(X)):
                X = X.copy()
                X[missing] = 0.
        
        # Convert to internally used data type
        missing = np.asarray(missing, dtype=BOOL, order='F')
        assert_all_finite(missing)
        if missing.ndim == 1:
            missing = missing[:, np.newaxis]
        
        return X, missing

    def _scrub(self, X, y, sample_weight, output_weight, missing, **kwargs):
        '''
        Sanitize input data.
        '''
        # Check for sparseness
        if sparse.issparse(y):
            raise TypeError('A sparse matrix was passed, but dense data '
                            'is required. Use y.toarray() to convert to '
                            'dense.')
        if sparse.issparse(sample_weight):
            raise TypeError('A sparse matrix was passed, but dense data '
                            'is required. Use sample_weight.toarray()'
                            'to convert to dense.')
        if sparse.issparse(output_weight):
            raise TypeError('A sparse matrix was passed, but dense data '
                            'is required. Use output_weight.toarray()'
                            'to convert to dense.')

        # Check whether X is the output of patsy.dmatrices
        if y is None and isinstance(X, tuple):
            y, X = X

        # Handle X separately
        X, missing = self._scrub_x(X, missing, **kwargs)

        # Convert y to internally used data type
        y = np.asarray(y, dtype=np.float64)
        assert_all_finite(y)

        if len(y.shape) == 1:
            y = y[:, np.newaxis]

        # Deal with sample_weight
        if sample_weight is None:
            sample_weight = np.ones((y.shape[0], 1), dtype=y.dtype)
        else:
            sample_weight = np.asarray(sample_weight, dtype=np.float64)
            assert_all_finite(sample_weight)
            if len(sample_weight.shape) == 1:
                sample_weight = sample_weight[:, np.newaxis]
        # Deal with output_weight
        if output_weight is not None:
            output_weight = np.asarray(output_weight, dtype=np.float64)
            assert_all_finite(output_weight)

        # Make sure dimensions match
        if y.shape[0] != X.shape[0]:
            raise ValueError('X and y do not have compatible dimensions.')
        if y.shape[0] != sample_weight.shape[0]:
            raise ValueError(
                'y and sample_weight do not have compatible dimensions.')
        if output_weight is not None and y.shape[1] != output_weight.shape[0]:
            raise ValueError(
                'y and output_weight do not have compatible dimensions.')
        if y.shape[1] > 1:
            if sample_weight.shape[1] == 1 and output_weight is not None:
                sample_weight = np.repeat(sample_weight, y.shape[1], axis=1)
        if output_weight is not None:
            sample_weight *= output_weight

        # Make sure everything is finite (except X, which is allowed to have
        # missing values)
        assert_all_finite(missing)
        assert_all_finite(y)
        assert_all_finite(sample_weight)
        assert_all_finite(output_weight)

        # Make sure everything is consistent
        check_X_y(X, y, accept_sparse=False, multi_output=True,
                  force_all_finite=False)

        return X, y, sample_weight, None, missing

    def fit(self, X, y=None,
            sample_weight=None,
            output_weight=None,
            missing=None,
            xlabels=None, linvars=[]):
        '''
        Fit an Earth model to the input data X and y.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features the training predictors.
            The X parameter can be a numpy array, a pandas DataFrame, a patsy
            DesignMatrix, or a tuple of patsy DesignMatrix objects as
            output by patsy.dmatrices.


        y : array-like, optional (default=None), shape = [m, p] where m is the
            number of samples The training response, p the number of outputs.
            The y parameter can be a numpy array, a pandas DataFrame,
            a Patsy DesignMatrix, or can be left as None (default) if X was
            the output of a call to patsy.dmatrices (in which case, X contains
            the response).


        sample_weight : array-like, optional (default=None), shape = [m]
             where m is the number of samples.
             Sample weights for training.  Weights must be greater than or
             equal to zero. Rows with zero weight do not contribute at all.
             Weights are useful when dealing with heteroscedasticity.
             In such cases, the weight should be proportional to the inverse of
             the (known) variance.

        output_weight : array-like, optional (default=None), shape = [p]
             where p is the number of outputs.
             Output weights for training. Weights must be greater than or equal
             to zero. Output with zero weight do not contribute at all.

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a  patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        linvars : iterable of strings or ints, optional (empty by default)
            Used to specify features that may only enter terms as linear basis
            functions (without knots).  Can include both column numbers and
            column names (see xlabels, below).  If left empty, some variables
            may still enter linearly during the forward pass if no knot would
            provide a reduction in GCV compared to the linear function.
            Note that this feature differs from the R package earth.


        xlabels : iterable of strings, optional (empty by default)
            The xlabels argument can be used to assign names to data columns.
            This argument is not generally needed, as names can be captured
            automatically from most standard data structures.
            If included, must have length n, where n is the number of features.
            Note that column order is used to compute term values and make
            predictions, not column names.


        '''
        # Format and label the data
        if xlabels is None:
            self.xlabels_ = self._scrape_labels(X)
        else:
            if len(xlabels) != X.shape[1]:
                raise ValueError('The length of xlabels is not the '
                                 'same as the number of columns of X')
            self.xlabels_ = xlabels
        if self.feature_importance_type is not None:
            feature_importance_type = self.feature_importance_type
            try:
                is_str = isinstance(feature_importance_type, basestring)
            except NameError:
                is_str = isinstance(feature_importance_type, str)
            if is_str:
                feature_importance_type = [feature_importance_type]
            for k in feature_importance_type:
                if k not in FEAT_IMP_CRITERIA:
                    msg = ("'{}' is not valid value for feature_importance, "
                           "allowed critera are : {}".format(k, FEAT_IMP_CRITERIA))
                    raise ValueError(msg)

            if len(feature_importance_type) > 0 and self.enable_pruning is False:
                raise ValueError("Cannot compute feature importances because pruning is disabled,"
                                 "please re-enable pruning by setting enable_pruning to True in order"
                                 "to enable feature importance estimation")

        self.linvars_ = linvars
        X, y, sample_weight, output_weight, missing = self._scrub(
            X, y, sample_weight, output_weight, missing)

        # Do the actual work
        self.forward_pass(X, y,
                          sample_weight, output_weight, missing,
                          self.xlabels_, linvars, skip_scrub=True)
        if self.enable_pruning is True:
            self.pruning_pass(X, y,
                              sample_weight, output_weight, missing,
                              skip_scrub=True)
        if hasattr(self, 'smooth') and self.smooth:
            self.basis_ = self.basis_.smooth(X)
        self.linear_fit(X, y, sample_weight, output_weight, missing,
                        skip_scrub=True)
        return self

#     def forward_pass2(self, X, y=None,
#                        sample_weight=None, output_weight=None,
#                        missing=None,
#                        xlabels=None, linvars=[]):
#         # Label and format data
#         if xlabels is None:
#             self.xlabels_ = self._scrape_labels(X)
#         else:
#             self.xlabels_ = xlabels
#         X, y, sample_weight, output_weight, missing = self._scrub(
#             X, y, sample_weight, output_weight, missing)
#
#         # Do the actual work
#         args = self._pull_forward_args(**self.__dict__)
#         forward_passer = ForwardPasser(
#             X, missing, y, sample_weight, output_weight,
#             xlabels=self.xlabels_, linvars=linvars, **args)
# #         forward_passer.run()
#         linvars_ = []
#         self.forward_pass_record_, self.basis_ = forward_pass(X, missing, y,
#             sample_weight, output_weight, xlabels=self.xlabels_, linvars=linvars)
# #         self.basis_ = forward_passer.get_basis()
#
    def forward_pass(self, X, y=None,
                     sample_weight=None, output_weight=None,
                     missing=None,
                     xlabels=None, linvars=[], skip_scrub=False):
        '''
        Perform the forward pass of the multivariate adaptive regression
        splines algorithm.  Users will normally want to call the fit method
        instead, which performs the forward pass, the pruning pass,
        and a linear fit to determine the final model coefficients.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples and n
            is the number of features The training predictors.
            The X parameter can be a numpy array, a pandas DataFrame, a patsy
            DesignMatrix, or a tuple of patsy DesignMatrix objects as output
            by patsy.dmatrices.

        y : array-like, optional (default=None), shape = [m, p] where m is the
            number of samples, p the number of outputs.
            The y parameter can be a numpy array, a pandas DataFrame,
            a Patsy DesignMatrix, or can be left as None (default) if X was
            the output of a call to patsy.dmatrices (in which case, X contains
            the response).

        sample_weight : array-like, optional (default=None), shape = [m]
             where m is the number of samples.
             Sample weights for training.  Weights must be greater than or
             equal to zero. Rows with zero weight do not contribute at all.
             Weights are useful when dealing with heteroscedasticity.  In such
             cases, the weight should be proportional to the inverse of the
             (known) variance.

        output_weight : array-like, optional (default=None), shape = [p]
             where p is the number of outputs.
             The total mean squared error (MSE) is a weighted sum of
             mean squared errors (MSE) associated to each output, where
             the weights are given by output_weight.
             Output weights must be greater than or equal
             to zero. Outputs with zero weight do not contribute at all
             to the total mean squared error (MSE).

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a  patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        linvars : iterable of strings or ints, optional (empty by default)
            Used to specify features that may only enter terms as linear basis
            functions (without knots).  Can include both column numbers and
            column names (see xlabels, below).


        xlabels : iterable of strings, optional (empty by default)
            The xlabels argument can be used to assign names to data columns.
            This argument is not generally needed, as names can be captured
            automatically from most standard data structures.
            If included, must have length n, where n is the number of features.
            Note that column order is used to compute term values and make
            predictions, not column names.


        '''
        # Label and format data
        if xlabels is None:
            self.xlabels_ = self._scrape_labels(X)
        else:
            self.xlabels_ = xlabels
        if not skip_scrub:
            X, y, sample_weight, output_weight, missing = self._scrub(
                X, y, sample_weight, output_weight, missing)

        # Do the actual work
        args = self._pull_forward_args(**self.__dict__)
        forward_passer = ForwardPasser(
            X, missing, y, sample_weight,
            xlabels=self.xlabels_, linvars=linvars, **args)
        forward_passer.run()
        self.forward_pass_record_ = forward_passer.trace()
        self.basis_ = forward_passer.get_basis()

    def pruning_pass(self, X, y=None, sample_weight=None, output_weight=None,
                     missing=None, skip_scrub=False):
        '''
        Perform the pruning pass of the multivariate adaptive regression
        splines algorithm.  Users will normally want to call the fit
        method instead, which performs the forward pass, the pruning
        pass, and a linear fit to determine the final model coefficients.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features The training predictors.
            The X parameter can be a numpy array, a pandas DataFrame, a patsy
            DesignMatrix, or a tuple of patsy DesignMatrix objects as output
            by patsy.dmatrices.

        y : array-like, optional (default=None), shape = [m, p] where m is the
            number of samples, p the number of outputs.
            The y parameter can be a numpy array, a pandas DataFrame,
            a Patsy DesignMatrix, or can be left as None (default) if X was
            the output of a call to patsy.dmatrices (in which case, X contains
            the response).

        sample_weight : array-like, optional (default=None), shape = [m]
             where m is the number of samples.
             Sample weights for training.  Weights must be greater than or
             equal to zero. Rows with zero weight do not contribute at all.
             Weights are useful when dealing with heteroscedasticity.  In such
             cases, the weight should be proportional to the inverse of the
             (known) variance.

        output_weight : array-like, optional (default=None), shape = [p]
             where p is the number of outputs.
             The total mean squared error (MSE) is a weighted sum of
             mean squared errors (MSE) associated to each output, where
             the weights are given by output_weight.
             Output weights must be greater than or equal
             to zero. Outputs with zero weight do not contribute at all
             to the total mean squared error (MSE).

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a  patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        '''
        # Format data
        if not skip_scrub:
            X, y, sample_weight, output_weight, missing = self._scrub(
                X, y, sample_weight, output_weight, missing)

#         if sample_weight.shape[1] == 1 and y.shape[1] != 1:
#             sample_weight = np.repeat(sample_weight,y.shape[1],axis=1)

        # Pull arguments from self
        args = self._pull_pruning_args(**self.__dict__)

        # Do the actual work
        pruning_passer = PruningPasser(
            self.basis_, X, missing, y, sample_weight,
            **args)
        pruning_passer.run()

        imp = pruning_passer.feature_importance
        self._feature_importances_dict = imp
        if len(imp) == 1: # if only one criterion then return it only
            imp = imp[list(imp.keys())[0]]
        elif len(imp) == 0:
            imp = None
        self.feature_importances_ = imp
        self.pruning_pass_record_ = pruning_passer.trace()


#         # Format data
#         X, y, sample_weight, output_weight, missing = self._scrub(
#             X, y, sample_weight, output_weight, missing)
#
#         # Dimensions
#         m, n = X.shape
#         basis_size = len(self.basis_)
#         pruned_basis_size = self.basis_.plen()
#
#         self.pruning_pass_record_ = PruningPassRecord(m, n, args['penalty'], sst, pruned_basis_size)
#
#
#
#         # Pull arguments from self
#         args = self._pull_pruning_args(**self.__dict__)
#         penalty = args.get('penalty', 3.0)
#
#         # Compute prune sets
#         prune_sets = [[basis_size - i for i in range(n)] for n in range(basis_size)]
#
#         # Create the record object
#         record = PruningPassRecord('gcv', basis_size)
#
#         # Score each prune set to find the best
#         best_prune_set = []
#         best_score = float('inf')
#         for prune_set in prune_sets:
#             print 1
#             print prune_set
#             sys.stdout.flush()
#             # Prune this prune set
#             for idx in prune_set:
#                 self.basis_[idx].prune()
#
#             print 2
#             sys.stdout.flush()
#             # Score this prune set
#             self.linear_fit(X, y, sample_weight, output_weight, missing, skip_scrub=True)
#             y_pred = self.predict(X, missing, skip_scrub=True)
#             score = gcv(np.mean(((y - y_pred) * np.sqrt(sample_weight)) ** 2),
#                         basis_size - len(prune_set), m, penalty)
#             r2 = self.score(X, y, sample_weight, output_weight, missing, skip_scrub=True)
#
#             print 3
#             sys.stdout.flush()
#             # Minimizer
#             if score < best_score:
#                 best_score = score
#                 best_prune_set = prune_set
#
#             print 4
#             sys.stdout.flush()
#             # Unprune for next iteration
#             for idx in prune_set:
#                 self.basis_[idx].unprune()
#
#             print 5
#             sys.stdout.flush()
#             # Add to the record
#             record.add(prune_set, score, r2)
#
#         # Apply the best prune set
#         for idx in best_prune_set:
#             self.basis_[idx].prune()
#
#         self.pruning_pass_record_ = record

    def forward_trace(self):
        '''Return information about the forward pass.'''
        try:
            return self.forward_pass_record_
        except AttributeError:
            return None

    def pruning_trace(self):
        '''Return information about the pruning pass.'''
        try:
            return self.pruning_pass_record_
        except AttributeError:
            return None

    def trace(self):
        '''Return information about the forward and pruning passes.'''
        return EarthTrace(self.forward_trace(), self.pruning_trace())

    def summary(self):
        '''Return a string describing the model.'''
        result = ''
        if self.forward_trace() is None:
            result += 'Untrained Earth Model'
            return result
        elif self.pruning_trace() is None:
            result += 'Unpruned Earth Model\n'
        else:
            result += 'Earth Model\n'
        header = ['Basis Function', 'Pruned']
        if self.coef_.shape[0] > 1:
            header += ['Coefficient %d' %
                       i for i in range(self.coef_.shape[0])]
        else:
            header += ['Coefficient']
        data = []

        i = 0
        for bf in self.basis_:
            data.append([str(bf), 'Yes' if bf.is_pruned() else 'No'] + [
                          '%g' % self.coef_[c, i] if not bf.is_pruned() else
                          'None' for c in range(self.coef_.shape[0])])
            if not bf.is_pruned():
                i += 1
        result += ascii_table(header, data)
        result += '\n'
        result += 'MSE: %.4f, GCV: %.4f, RSQ: %.4f, GRSQ: %.4f' % (
            self.mse_, self.gcv_, self.rsq_, self.grsq_)
        return result

    def summary_feature_importances(self, sort_by=None):
        """
        Returns a string containing a printable summary of the estimated
        feature importances.

        Parameters
        ----------
        sory_by : string, optional
            it refers to a feature importance type name : 'gcv', 'rss'
            or 'nb_subsets'.
            In case it is provided, the features are sorted
            according to the feature importance type corresponding
            to `sort_by`. In case it is not provided, the features are
            not sorted.
        """
        result = ''
        if self._feature_importances_dict:
            max_label_length = max(map(len, self.xlabels_)) + 5
            result += (max_label_length * ' ' +
                       '    '.join(self._feature_importances_dict.keys()) + '\n')
            labels = np.array(self.xlabels_)
            if sort_by:
                if sort_by not in self._feature_importances_dict.keys():
                    raise ValueError('Invalid feature importance type name '
                                     'to sort with : %s, available : %s' % (
                                         sort_by,
                                         self._feature_importances_dict.keys()))
                imp = self._feature_importances_dict[sort_by]
                indices = np.argsort(imp)[::-1]
            else:
                indices = np.arange(len(labels))
            labels = labels[indices]
            for i, label in enumerate(labels):
                result += label + ' ' * (max_label_length - len(label))
                for crit_name, imp in self._feature_importances_dict.items():
                    imp = imp[indices]
                    result += '%.2f' % imp[i] + (len(crit_name) ) * ' '
                result += '\n'
        return result

    def linear_fit(self, X, y=None, sample_weight=None, output_weight=None,
                   missing=None, skip_scrub=False):
        '''
        Solve the linear least squares problem to determine the coefficients
        of the unpruned basis functions.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples and n
            is the number of features The training predictors.  The X parameter
            can be a numpy array, a pandas DataFrame, a patsy
            DesignMatrix, or a tuple of patsy DesignMatrix objects as output
            by patsy.dmatrices.

        y : array-like, optional (default=None), shape = [m, p] where m is the
            number of samples, p the number of outputs.
            The y parameter can be a numpy array, a pandas DataFrame,
            a Patsy DesignMatrix, or can be left as None (default) if X was
            the output of a call to patsy.dmatrices (in which case, X contains
            the response).

        sample_weight : array-like, optional (default=None), shape = [m]
             where m is the number of samples.
             Sample weights for training.  Weights must be greater than or
             equal to zero. Rows with zero weight do not contribute at all.
             Weights are useful when dealing with heteroscedasticity.  In such
             cases, the weight should be proportional to the inverse of the
             (known) variance.

        output_weight : array-like, optional (default=None), shape = [p]
             where p is the number of outputs.
             The total mean squared error (MSE) is a weighted sum of
             mean squared errors (MSE) associated to each output, where
             the weights are given by output_weight.
             Output weights must be greater than or equal
             to zero. Outputs with zero weight do not contribute at all
             to the total mean squared error (MSE).

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a  patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        '''
        # Format data
        if not skip_scrub:
            X, y, sample_weight, output_weight, missing = self._scrub(
                X, y, sample_weight, output_weight, missing)

#         if sample_weight.shape[1]:
#             sample_weight = np.repeat(sample_weight,y.shape[1],axis=1)

        # Solve the linear least squares problem
        self.coef_ = []
        resid_ = []
        total_weight = 0.
        mse0 = 0.
        for i in range(y.shape[1]):

            # Figure out the weight column
            if sample_weight.shape[1] > 1:
                w = sample_weight[:, i]
            else:
                w = sample_weight[:, 0]

            # Transform into basis space
            B = self.transform(X, missing)  # * w[:, None]
            apply_weights_2d(B, w)

            # Compute total weight
            total_weight += np.sum(w)

            # Apply weights to y
            weighted_y = y.copy()
            weighted_y *= np.sqrt(w[:, np.newaxis])

            # Compute the mse0
            mse0 += np.sum((weighted_y[:, i] -
                            np.average(weighted_y[:, i])) ** 2)

            coef, resid = np.linalg.lstsq(B, weighted_y[:, i])[0:2]
            self.coef_.append(coef)
            if not resid:
                resid = np.array(
                    [np.sum((np.dot(B, coef) - weighted_y[:, i]) ** 2)])
            resid_.append(resid)
        resid_ = np.array(resid_)
        self.coef_ = np.array(self.coef_)
        # Compute the final mse, gcv, rsq, and grsq (may be different from the
        # pruning scores if the model has been smoothed)
        self.mse_ = np.sum(resid_) / total_weight
        mse0 = mse0 / total_weight
        self.gcv_ = gcv(self.mse_,
                        coef.shape[0], X.shape[0],
                        self.get_penalty())
        gcv0 = gcv(mse0,
                   1, X.shape[0],
                   self.get_penalty())
        if mse0 != 0.:
            self.rsq_ = 1.0 - (self.mse_ / mse0)
        else:
            self.rsq_ = 1.0
        if gcv0 != 0.:
            self.grsq_ = 1.0 - (self.gcv_ / gcv0)
        else:
            self.grsq_ = 1.0

#
#
#
#         for p, coef in enumerate(self.coef_):
#             mse_p = resid_[p].sum() / float(X.shape[0])
#             gcv_[p] = gcv(mse_p,
#                           coef.shape[0], X.shape[0],
#                           self.get_penalty())
#             self.gcv_ += gcv_[p] * output_weight[p]
#         y_avg = np.average(y, weights=sample_weight if
#                            sample_weight.shape == y.shape else sample_weight.flatten(), axis=0)
#         y_sqr = (y - y_avg[np.newaxis, :]) ** 2
#
#         rsq_ = ((1 - resid_.sum(axis=1) / y_sqr.sum(axis=0)) * output_weight)
#         self.rsq_ = rsq_.sum()
#         gcv0 = np.empty(y.shape[1])
#         for p in range(y.shape[1]):
#             mse0_p = (y_sqr[:, p].sum()) / float(X.shape[0])
#             gcv0[p] = gcv(mse0_p, 1, X.shape[0], self.get_penalty())
#         self.grsq_ = ((1 - (gcv_ / gcv0)) * output_weight).sum()

    def predict(self, X, missing=None, skip_scrub=False):
        '''
        Predict the response based on the input data X.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples and n
            is the number of features
            The training predictors.  The X parameter can be a numpy
            array, a pandas DataFrame, or a patsy DesignMatrix.

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a  patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

       Returns
       -------
            y : array of shape = [m] or [m, p] where m is the number of samples
                and p is the number of outputs
                The predicted values.
        '''
        if not skip_scrub:
            X, missing = self._scrub_x(X, missing)
        B = self.transform(X, missing)
        y = np.dot(B, self.coef_.T)
        if y.shape[1] == 1:
            return y[:, 0]
        else:
            return y

    def predict_deriv(self, X, variables=None, missing=None):
        '''
        Predict the first derivatives of the response based on the input
        data X.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples and n
            is the number of features The training predictors. The X parameter
            can be a numpy array, a pandas DataFrame, or a patsy DesignMatrix.

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        variables : list
            The variables over which derivatives will be computed.  Each column
            in the resulting array corresponds to a variable.  If not
            specified, all variables are used (even if some are not relevant
            to the final model and have derivatives that are identically zero).

       Returns
       -------

       X_deriv : array of shape = [m, n, p] where m is the number of samples, n
                 is the number of features if 'variables' is not specified
                 otherwise it is len(variables) and p is the number of outputs.
                 For each sample, X_deriv represents the first derivative of
                 each response  with respect to each variable.

        '''

        check_is_fitted(self, "basis_")

        if type(variables) in (str, int):
            variables = [variables]
        if variables is None:
            variables_of_interest = list(range(len(self.xlabels_)))
        else:
            variables_of_interest = []
            for var in variables:
                if isinstance(var, int):
                    variables_of_interest.append(var)
                else:
                    variables_of_interest.append(self.xlabels_.index(var))
        X, missing = self._scrub_x(X, missing)
        J = np.zeros(shape=(X.shape[0],
                            len(variables_of_interest),
                            self.coef_.shape[0]))
        b = np.empty(shape=X.shape[0])
        j = np.empty(shape=X.shape[0])
        self.basis_.transform_deriv(
            X, missing, b, j, self.coef_, J, variables_of_interest, True)
        return J

    def score(self, X, y=None, sample_weight=None, output_weight=None,
              missing=None, skip_scrub=False):
        '''
        Calculate the generalized r^2 of the model on data X and y.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features The training predictors.
            The X parameter can be a numpy array, a pandas DataFrame, a patsy
            DesignMatrix, or a tuple of patsy DesignMatrix objects as output
            by patsy.dmatrices.

        y : array-like, optional (default=None), shape = [m, p] where m is the
            number of samples, p the number of outputs.
            The y parameter can be a numpy array, a pandas DataFrame,
            a Patsy DesignMatrix, or can be left as None (default) if X was
            the output of a call to patsy.dmatrices (in which case, X contains
            the response).

        sample_weight : array-like, optional (default=None), shape = [m]
             where m is the number of samples.
             Sample weights for training.  Weights must be greater than or
             equal to zero. Rows with zero weight do not contribute at all.
             Weights are useful when dealing with heteroscedasticity.  In such
             cases, the weight should be proportional to the inverse of the
             (known) variance.

        output_weight : array-like, optional (default=None), shape = [p]
             where p is the number of outputs.
             The total mean squared error (MSE) is a weighted sum of
             mean squared errors (MSE) associated to each output, where
             the weights are given by output_weight.
             Output weights must be greater than or equal
             to zero. Outputs with zero weight do not contribute at all
             to the total mean squared error (MSE).

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        Returns
        -------

        score : float with a maximum value of 1 (it can be negative). The score
                is the generalized r^2 of the model on data X and y, the higher
                the score the better the fit is.

        '''
        check_is_fitted(self, "basis_")
        if not skip_scrub:
            X, y, sample_weight, output_weight, missing = self._scrub(
                X, y, sample_weight, output_weight, missing)
        if sample_weight.shape[1] == 1 and y.shape[1] > 1:
            sample_weight = np.repeat(sample_weight, y.shape[1], axis=1)
        y_hat = self.predict(X)
        if len(y_hat.shape) == 1:
            y_hat = y_hat[:, None]

        residual = y - y_hat
#         total_weight = np.sum(sample_weight)
        mse = np.sum(sample_weight * (residual ** 2))
        y_avg = np.average(y, weights=sample_weight, axis=0)

        mse0 = np.sum(sample_weight * ((y - y_avg) ** 2))
#         mse0 = np.sum(y_sqr * output_weight) / m
        return 1 - (mse / mse0)

    def score_samples(self, X, y=None, missing=None):
        '''

        Calculate sample-wise fit scores.

        Parameters
        ----------

        X : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features The training predictors.
            The X parameter can be a numpy array, a pandas DataFrame, a patsy
            DesignMatrix, or a tuple of patsy DesignMatrix objects as output
            by patsy.dmatrices.

        y : array-like, optional (default=None), shape = [m, p] where m is the
            number of samples, p the number of outputs.
            The y parameter can be a numpy array, a pandas DataFrame,
            a Patsy DesignMatrix, or can be left as None (default) if X was
            the output of a call to patsy.dmatrices (in which case, X contains
            the response).

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a  patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        Returns
        -------

        scores : array of shape=[m, p] of floats with maximum value of 1
                 (it can be negative).
                 The scores represent how good each output of each example is
                 predicted, a perfect score would be 1
                 (the score can be negative).

        '''
        X, y, sample_weight, output_weight, missing = self._scrub(
            X, y, None, None, missing)
        y_hat = self.predict(X, missing=missing)
        residual = 1 - (y - y_hat) ** 2 / y**2
        return residual

    def transform(self, X, missing=None):
        '''
        Transform X into the basis space.  Normally, users will call the
        predict method instead, which both transforms into basis space
        calculates the weighted sum of basis terms to produce a prediction
        of the response.  Users may wish to call transform directly in some
        cases.  For example, users may wish to apply other statistical or
        machine learning algorithms, such as generalized linear regression,
        in basis space.


        Parameters
        ----------
        X : array-like, shape = [m, n] where m is the number of samples and n
            is the number of features
            The training predictors.  The X parameter can be a numpy array, a
            pandas DataFrame, or a patsy DesignMatrix.

        missing : array-like, shape = [m, n] where m is the number of samples
            and n is the number of features.
            The missing parameter can be a numpy array, a pandas DataFrame, or
            a patsy DesignMatrix.  All entries will be interpreted as boolean
            values, with True indicating the corresponding entry in X should be
            interpreted as missing.  If the missing argument not used but the X
            argument is a pandas DataFrame, missing will be inferred from X if
            allow_missing is True.

        Returns
        -------

        B: array of shape [m, nb_terms] where m is the number of samples and
           nb_terms is the number of terms (or basis functions) obtained after
           fitting (which is the number of elements of the attribute `basis_`).
           B represents the values of the basis functions evaluated at each
           sample.
        '''

        check_is_fitted(self, "basis_")
        X, missing = self._scrub_x(X, missing)
        B = np.empty(shape=(X.shape[0], self.basis_.plen()), order='F')
        self.basis_.transform(X, missing, B)
        return B

    def get_penalty(self):
        '''Get the penalty parameter being used.  Default is 3.'''
        if 'penalty' in self.__dict__ and self.penalty is not None:
            return self.penalty
        else:
            return 3.0


class EarthTrace(object):

    def __init__(self, forward_trace, pruning_trace):
        self.forward_trace = forward_trace
        self.pruning_trace = pruning_trace

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.forward_trace == other.forward_trace and
                self.pruning_trace == other.pruning_trace)

    def __str__(self):
        result = ''
        result += 'Forward Pass\n'
        result += str(self.forward_trace)
        result += '\n'
        result += self.forward_trace.final_str()
        result += '\n\n'
        result += 'Pruning Pass\n'
        result += str(self.pruning_trace)
        result += '\n'
        result += self.pruning_trace.final_str()
        result += '\n'
        return result



'''
Created on Feb 24, 2013

@author: jasonrudy
'''
import pickle
import copy
import os
from .testing_utils import (if_statsmodels, if_pandas, if_patsy,
                            if_environ_has, assert_list_almost_equal_value,
                            assert_list_almost_equal,
                            if_sklearn_version_greater_than_or_equal_to,
                            if_platform_not_win_32)
from nose.tools import (assert_equal, assert_true, assert_almost_equal,
                        assert_list_equal, assert_raises, assert_not_equal)
import numpy
from scipy.sparse import csr_matrix
from pyearth._types import BOOL
from pyearth._basis import (Basis, ConstantBasisFunction,
                            HingeBasisFunction, LinearBasisFunction)
from pyearth import Earth
import pyearth
from numpy.testing.utils import assert_array_almost_equal

regenerate_target_files = False

numpy.random.seed(1)
basis = Basis(10)
constant = ConstantBasisFunction()
basis.append(constant)
bf1 = HingeBasisFunction(constant, 0.1, 10, 1, False, 'x1')
bf2 = HingeBasisFunction(constant, 0.1, 10, 1, True, 'x1')
bf3 = LinearBasisFunction(bf1, 2, 'x2')
basis.append(bf1)
basis.append(bf2)
basis.append(bf3)
X = numpy.random.normal(size=(1000, 10))
missing = numpy.zeros_like(X, dtype=BOOL)
B = numpy.empty(shape=(1000, 4), dtype=numpy.float64)
basis.transform(X, missing, B)
beta = numpy.random.normal(size=4)
y = numpy.empty(shape=1000, dtype=numpy.float64)
y[:] = numpy.dot(B, beta) + numpy.random.normal(size=1000)
default_params = {"penalty": 1}

@if_platform_not_win_32
@if_sklearn_version_greater_than_or_equal_to('0.17.2')
def test_check_estimator():
    numpy.random.seed(0)
    import sklearn.utils.estimator_checks
    sklearn.utils.estimator_checks.MULTI_OUTPUT.append('Earth')
    sklearn.utils.estimator_checks.check_estimator(Earth)


def test_get_params():
    assert_equal(
        Earth().get_params(), {'penalty': None, 'min_search_points': None,
                               'endspan_alpha': None, 'check_every': None,
                               'max_terms': None, 'max_degree': None,
                               'minspan_alpha': None, 'thresh': None,
                               'zero_tol': None,
                               'minspan': None, 'endspan': None,
                               'allow_linear': None,
                               'use_fast': None, 'fast_K': None,
                               'fast_h': None, 'smooth': None,
                               'enable_pruning': True,
                               'allow_missing': False,
                               'feature_importance_type': None,
                               'verbose': False})
    assert_equal(
        Earth(
            max_degree=3).get_params(), {'penalty': None,
                                         'min_search_points': None,
                                         'endspan_alpha': None,
                                         'check_every': None,
                                         'max_terms': None, 'max_degree': 3,
                                         'minspan_alpha': None,
                                         'thresh': None, 'zero_tol': None,
                                         'minspan': None,
                                         'endspan': None,
                                         'allow_linear': None,
                                         'use_fast': None,
                                         'fast_K': None, 'fast_h': None,
                                         'smooth': None,
                                         'enable_pruning': True,
                                         'allow_missing': False,
                                         'feature_importance_type': None,
                                         'verbose': False})


@if_statsmodels
def test_linear_fit():
    from statsmodels.regression.linear_model import GLS, OLS

    earth = Earth(**default_params)
    earth.fit(X, y)
    earth.linear_fit(X, y)
    soln = OLS(y, earth.transform(X)).fit().params
    assert_almost_equal(numpy.mean((earth.coef_ - soln) ** 2), 0.0)

    sample_weight = 1.0 / (numpy.random.normal(size=y.shape) ** 2)
    earth.fit(X, y)
    earth.linear_fit(X, y, sample_weight)
    soln = GLS(y, earth.transform(
        X), 1.0 / sample_weight).fit().params
    assert_almost_equal(numpy.mean((earth.coef_ - soln) ** 2), 0.0)


def test_sample_weight():
    group = numpy.random.binomial(1, .5, size=1000) == 1
    sample_weight = 1 / (group * 100 + 1.0)
    x = numpy.random.uniform(-10, 10, size=1000)
    y = numpy.abs(x)
    y[group] = numpy.abs(x[group] - 5)
    y += numpy.random.normal(0, 1, size=1000)
    model = Earth().fit(x[:, numpy.newaxis], y, sample_weight=sample_weight)

    # Check that the model fits better for the more heavily weighted group
    assert_true(model.score(x[group], y[group]) < model.score(
        x[numpy.logical_not(group)], y[numpy.logical_not(group)]))

    # Make sure that the score function gives the same answer as the trace
    pruning_trace = model.pruning_trace()
    rsq_trace = pruning_trace.rsq(model.pruning_trace().get_selected())
    assert_almost_equal(model.score(x, y, sample_weight=sample_weight),
                        rsq_trace)

    # Uncomment below to see what this test situation looks like
#     from matplotlib import pyplot
#     print model.summary()
#     print model.score(x,y,sample_weight = sample_weight)
#     pyplot.figure()
#     pyplot.plot(x,y,'b.')
#     pyplot.plot(x,model.predict(x),'r.')
#     pyplot.show()


def test_output_weight():
    x = numpy.random.uniform(-1, 1, size=(1000, 1))
    y = (numpy.dot(x, numpy.random.normal(0, 1, size=(1, 10)))) ** 5 + 1
    y = (y - y.mean(axis=0)) / y.std(axis=0)
    group = numpy.array([1] * 5 + [0] * 5)
    output_weight = numpy.array([1] * 5 + [2] * 5, dtype=float)
    model = Earth().fit(x, y, output_weight=output_weight)

    # Check that the model fits at least better
    # the more heavily weighted group
    mse = ((model.predict(x) - y)**2).mean(axis=0)
    group1_mean = mse[group].mean()
    group2_mean = mse[numpy.logical_not(group)].mean()
    assert_true(group1_mean > group2_mean or
                round(abs(group1_mean - group2_mean), 7) == 0)


def test_missing_data():
    numpy.random.seed(0)
    earth = Earth(allow_missing=True, **default_params)
    missing_ = numpy.random.binomial(1, .05, X.shape).astype(bool)
    X_ = X.copy()
    X_[missing_] = None
    earth.fit(X_, y)
    res = str(earth.score(X_, y))
    filename = os.path.join(os.path.dirname(__file__),
                            'earth_regress_missing_data.txt')
    if regenerate_target_files:
        with open(filename, 'w') as fl:
            fl.write(res)
    with open(filename, 'r') as fl:
        prev = fl.read()
    try:
        assert_true(abs(float(res) - float(prev)) < .03)
    except AssertionError:
        print('Got %f, %f' % (float(res), float(prev)))
        raise

def test_fit():
    numpy.random.seed(0)
    earth = Earth(**default_params)
    earth.fit(X, y)
    res = str(earth.rsq_)
    filename = os.path.join(os.path.dirname(__file__),
                            'earth_regress.txt')
    if regenerate_target_files:
        with open(filename, 'w') as fl:
            fl.write(res)
    with open(filename, 'r') as fl:
        prev = fl.read()
    assert_true(abs(float(res) - float(prev)) < .05)


def test_smooth():
    numpy.random.seed(0)
    model = Earth(penalty=1, smooth=True)
    model.fit(X, y)
    res = str(model.rsq_)
    filename = os.path.join(os.path.dirname(__file__),
                            'earth_regress_smooth.txt')
    if regenerate_target_files:
        with open(filename, 'w') as fl:
            fl.write(res)
    with open(filename, 'r') as fl:
        prev = fl.read()
    assert_true(abs(float(res) - float(prev)) < .05)


def test_linvars():
    earth = Earth(**default_params)
    earth.fit(X, y, linvars=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    res = str(earth.rsq_)
    filename = os.path.join(os.path.dirname(__file__),
                            'earth_linvars_regress.txt')
    if regenerate_target_files:
        with open(filename, 'w') as fl:
            fl.write(res)
    with open(filename, 'r') as fl:
        prev = fl.read()

    assert_equal(res, prev)


def test_linvars_coefs():
    nb_vars = 11
    coefs = numpy.random.uniform(size=(nb_vars,))
    X = numpy.random.uniform(size=(100, nb_vars))
    bias = 1
    y = numpy.dot(X, coefs[:, numpy.newaxis]) + bias
    earth = Earth(max_terms=nb_vars * 2,
                  max_degree=1,
                  enable_pruning=False,
                  check_every=1,
                  thresh=0,
                  minspan=1,
                  endspan=1).fit(X, y, linvars=range(nb_vars))
    earth_bias = earth.coef_[0, 0]
    earth_coefs = sorted(earth.coef_[1:])

    assert_almost_equal(earth_bias, bias)
    assert_list_almost_equal(earth_coefs, sorted(coefs))


def test_score():
    earth = Earth(**default_params)
    model = earth.fit(X, y)
    record = model.pruning_trace()
    rsq = record.rsq(record.get_selected())
    assert_almost_equal(rsq, model.score(X, y))


@if_pandas
@if_environ_has('test_pathological_cases')
def test_pathological_cases():
    import pandas
    directory = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'pathological_data')
    cases = {'issue_44': {},
             'issue_50': {'penalty': 0.5,
                          'minspan': 1,
                          'allow_linear': False,
                          'endspan': 1,
                          'check_every': 1,
                          'sample_weight': 'issue_50_weight.csv'}}
    for case, settings in cases.iteritems():
        data = pandas.read_csv(os.path.join(directory, case + '.csv'))
        y = data['y']
        del data['y']
        X = data
        if 'sample_weight' in settings:
            filename = os.path.join(directory, settings['sample_weight'])
            sample_weight = pandas.read_csv(filename)['sample_weight']
            del settings['sample_weight']
        else:
            sample_weight = None
        model = Earth(**settings)
        model.fit(X, y, sample_weight=sample_weight)
        with open(os.path.join(directory, case + '.txt'), 'r') as infile:
            correct = infile.read()
        assert_equal(model.summary(), correct)


@if_pandas
def test_pandas_compatibility():
    import pandas
    X_df = pandas.DataFrame(X)
    y_df = pandas.DataFrame(y)
    colnames = ['xx' + str(i) for i in range(X.shape[1])]
    X_df.columns = colnames

    earth = Earth(**default_params)
    model = earth.fit(X_df, y_df)
    assert_list_equal(
        colnames, model.forward_trace()._getstate()['xlabels'])


@if_patsy
@if_pandas
def test_patsy_compatibility():
    import pandas
    import patsy
    X_df = pandas.DataFrame(X)
    y_df = pandas.DataFrame(y)
    colnames = ['xx' + str(i) for i in range(X.shape[1])]
    X_df.columns = colnames
    X_df['y'] = y
    y_df, X_df = patsy.dmatrices(
        'y ~ xx0 + xx1 + xx2 + xx3 + xx4 + xx5 + xx6 + xx7 + xx8 + xx9 - 1',
        data=X_df)

    model = Earth(**default_params).fit(X_df, y_df)
    assert_list_equal(
        colnames, model.forward_trace()._getstate()['xlabels'])


def test_pickle_compatibility():
    earth = Earth(**default_params)
    model = earth.fit(X, y)
    model_copy = pickle.loads(pickle.dumps(model))
    assert_true(model_copy == model)
    assert_array_almost_equal(model.predict(X), model_copy.predict(X))
    assert_true(model.basis_[0] is model.basis_[1]._get_root())
    assert_true(model_copy.basis_[0] is model_copy.basis_[1]._get_root())


def test_pickle_version_storage():
    earth = Earth(**default_params)
    model = earth.fit(X, y)
    assert_equal(model._version, pyearth.__version__)
    model._version = 'hello'
    assert_equal(model._version,'hello')
    model_copy = pickle.loads(pickle.dumps(model))
    assert_equal(model_copy._version, model._version)


def test_copy_compatibility():
    numpy.random.seed(0)
    model = Earth(**default_params).fit(X, y)
    model_copy = copy.copy(model)
    assert_true(model_copy == model)
    assert_array_almost_equal(model.predict(X), model_copy.predict(X))
    assert_true(model.basis_[0] is model.basis_[1]._get_root())
    assert_true(model_copy.basis_[0] is model_copy.basis_[1]._get_root())


def test_exhaustive_search():
    model = Earth(max_terms=13,
                  enable_pruning=False,
                  check_every=1,
                  thresh=0,
                  minspan=1,
                  endspan=1)
    model.fit(X, y)
    assert_equal(model.basis_.plen(), model.coef_.shape[1])
    assert_equal(model.transform(X).shape[1], len(model.basis_))


def test_nb_terms():

    for max_terms in (1, 3, 12, 13):
        model = Earth(max_terms=max_terms)
        model.fit(X, y)
        assert_true(len(model.basis_) <= max_terms + 2)
        assert_true(len(model.coef_) <= len(model.basis_))
        assert_true(len(model.coef_) >= 1)
        if max_terms == 1:
            assert_list_almost_equal_value(model.predict(X), y.mean())


def test_nb_degrees():
    for max_degree in (1, 2, 12, 13):
        model = Earth(max_terms=10,
                      max_degree=max_degree,
                      enable_pruning=False,
                      check_every=1,
                      thresh=0,
                      minspan=1,
                      endspan=1)
        model.fit(X, y)
        for basis in model.basis_:
            assert_true(basis.degree() >= 0)
            assert_true(basis.degree() <= max_degree)


def test_eq():
    model1 = Earth(**default_params)
    model2 = Earth(**default_params)
    assert_equal(model1, model2)
    assert_not_equal(model1, 5)

    params = {}
    params.update(default_params)
    params["penalty"] = 15
    model2 = Earth(**params)
    assert_not_equal(model1, model2)

    model3 = Earth(**default_params)
    model3.unknown_parameter = 5
    assert_not_equal(model1, model3)


def test_sparse():
    X_sparse = csr_matrix(X)

    model = Earth(**default_params)
    assert_raises(TypeError, model.fit, X_sparse, y)

    model = Earth(**default_params)
    model.fit(X, y)
    assert_raises(TypeError, model.predict, X_sparse)
    assert_raises(TypeError, model.predict_deriv, X_sparse)
    assert_raises(TypeError, model.transform, X_sparse)
    assert_raises(TypeError, model.score, X_sparse)

    model = Earth(**default_params)
    sample_weight = csr_matrix([1.] * X.shape[0])
    assert_raises(TypeError, model.fit, X, y, sample_weight)


def test_shape():
    model = Earth(**default_params)
    model.fit(X, y)

    X_reduced = X[:, 0:5]
    assert_raises(ValueError, model.predict, X_reduced)
    assert_raises(ValueError, model.predict_deriv, X_reduced)
    assert_raises(ValueError, model.transform, X_reduced)
    assert_raises(ValueError, model.score, X_reduced)

    model = Earth(**default_params)
    X_subsampled = X[0:10]
    assert_raises(ValueError, model.fit, X_subsampled, y)

    model = Earth(**default_params)
    y_subsampled = X[0:10]
    assert_raises(ValueError, model.fit, X, y_subsampled)

    model = Earth(**default_params)
    sample_weights = numpy.array([1.] * len(X))
    sample_weights_subsampled = sample_weights[0:10]
    assert_raises(ValueError, model.fit, X, y, sample_weights_subsampled)


def test_deriv():

    model = Earth(**default_params)
    model.fit(X, y)
    assert_equal(X.shape + (1,), model.predict_deriv(X).shape)
    assert_equal((X.shape[0], 1, 1), model.predict_deriv(X, variables=0).shape)
    assert_equal((X.shape[0], 1, 1), model.predict_deriv(
        X, variables='x0').shape)
    assert_equal((X.shape[0], 3, 1),
                 model.predict_deriv(X, variables=[1, 5, 7]).shape)
    assert_equal((X.shape[0], 0, 1),
                 model.predict_deriv(X, variables=[]).shape)

    res_deriv = model.predict_deriv(X, variables=['x2', 'x7', 'x0', 'x1'])
    assert_equal((X.shape[0], 4, 1), res_deriv.shape)

    res_deriv = model.predict_deriv(X, variables=['x0'])
    assert_equal((X.shape[0], 1, 1), res_deriv.shape)

    assert_equal((X.shape[0], 1, 1),
                 model.predict_deriv(X, variables=[0]).shape)


def test_xlabels():

    model = Earth(**default_params)
    assert_raises(ValueError, model.fit, X[
                  :, 0:5], y, xlabels=['var1', 'var2'])

    model = Earth(**default_params)
    model.fit(X[:, 0:3], y, xlabels=['var1', 'var2', 'var3'])

    model = Earth(**default_params)
    model.fit(X[:, 0:3], y, xlabels=['var1', 'var2', 'var3'])


def test_untrained():
    # NotFittedError moved from utils.validation to exceptions
    # some time after 0.17.1
    try:
        from sklearn.exceptions import NotFittedError
    except ImportError:
        from sklearn.utils.validation import NotFittedError

    # Make sure calling methods that require a fitted Earth object
    # raises the appropriate exception when using a not yet fitted
    # Earth object
    model = Earth(**default_params)
    assert_raises(NotFittedError, model.predict, X)
    assert_raises(NotFittedError, model.transform, X)
    assert_raises(NotFittedError, model.predict_deriv, X)
    assert_raises(NotFittedError, model.score, X)

    # the following should be changed to raise NotFittedError
    assert_equal(model.forward_trace(), None)
    assert_equal(model.pruning_trace(), None)
    assert_equal(model.summary(), "Untrained Earth Model")


def test_fast():
    earth = Earth(max_terms=10,
                  max_degree=5,
                  **default_params)
    earth.fit(X, y)
    normal_summary = earth.summary()
    earth = Earth(use_fast=True,
                  max_terms=10,
                  max_degree=5,
                  fast_K=10,
                  fast_h=1,
                  **default_params)
    earth.fit(X, y)
    fast_summary = earth.summary()
    assert_equal(normal_summary, fast_summary)


def test_feature_importance():
    criteria = ('rss', 'gcv', 'nb_subsets')
    for imp in criteria:
        earth = Earth(feature_importance_type=imp, **default_params)
        earth.fit(X, y)
        assert len(earth.feature_importances_) == X.shape[1]
    earth = Earth(feature_importance_type=criteria, **default_params)
    earth.fit(X, y)
    assert type(earth.feature_importances_) == dict
    assert set(earth.feature_importances_.keys()) == set(criteria)
    for crit, val in earth .feature_importances_.items():
        assert len(val) == X.shape[1]

    assert_raises(
            ValueError,
            Earth(feature_importance_type='bad_name', **default_params).fit,
            X, y)

    earth = Earth(feature_importance_type=('rss',), **default_params)
    earth.fit(X, y)
    assert len(earth.feature_importances_) == X.shape[1]

    assert_raises(
            ValueError,
            Earth(feature_importance_type='rss', enable_pruning=False, **default_params).fit,
            X, y)



cimport numpy as cnp
from _types cimport FLOAT_t, INT_t, INDEX_t, BOOL_t

cdef FLOAT_t log2(FLOAT_t x)

cpdef apply_weights_2d(cnp.ndarray[FLOAT_t, ndim=2] B,
                       cnp.ndarray[FLOAT_t, ndim=1] weights)

cpdef apply_weights_slice(cnp.ndarray[FLOAT_t, ndim=2] B,
                          cnp.ndarray[FLOAT_t, ndim=1] weights, INDEX_t column)

cpdef apply_weights_1d(cnp.ndarray[FLOAT_t, ndim=1] y,
                       cnp.ndarray[FLOAT_t, ndim=1] weights)

cpdef FLOAT_t gcv(FLOAT_t mse,
                  FLOAT_t basis_size, FLOAT_t data_size,
                  FLOAT_t penalty)

cpdef FLOAT_t gcv_adjust(FLOAT_t basis_size, FLOAT_t data_size, FLOAT_t penalty)

cpdef str_pad(string, length)

cpdef ascii_table(header, data, print_header=?, print_footer=?)



import pickle
import os
from nose.tools import assert_true, assert_false, assert_equal

import numpy

from pyearth._basis import Basis, ConstantBasisFunction, HingeBasisFunction, \
    LinearBasisFunction, SmoothedHingeBasisFunction

numpy.random.seed(0)



def export_python_function(earth_model):
    """
    Exports model as a pure python function, with no numpy/scipy/sklearn dependencies.
    :param earth_model: Trained pyearth model
    :return: A function that accepts an iterator over examples, and returns an iterator over transformed examples
    """
    i = 0
    accessors = []
    for bf in earth_model.basis_:
        if not bf.is_pruned():
            accessors.append(bf.func_factory(earth_model.coef_[0, i]))
            i += 1

    def func(example_iterator):
        return [sum(accessor(row) for accessor in accessors) for row in example_iterator]
    return func


def export_python_string(earth_model, function_name="model"):
    """
    Exports model as a string that evaluates as python code, with no numpy/scipy/sklearn dependencies.
    :param earth_model: Trained pyearth model
    :param function_name: string, optional, will be the name of the function in the returned string
    :return: string, when executed (either by writing to a file, or using `exec`, will define a python
      function that accepts an iterator over examples, and returns an iterator over transformed examples
    """
    i = 0
    accessors = []
    for bf in earth_model.basis_:
        if not bf.is_pruned():
            accessors.append(bf.func_string_factory(earth_model.coef_[0, i]))
            i += 1

    return """def {:s}(example_iterator):
    accessors = [{:s}]
    for x in example_iterator:
        yield sum(accessor(x) for accessor in accessors)
    """.format(function_name, ",\n\t\t".join(accessors))

def export_sympy_term_expressions(earth_model):
    """
    Construct a list of sympy expressions for all non-pruned terms in the model.

    :param earth_model: Trained pyearth model
    :return: a list of sympy expressions representing terms in the model.  These
      expressions are the symbolic equivalent of the Earth.transform method.

    """
    from sympy import Symbol, Add, Mul, Max, RealNumber, Piecewise, Pow, And, nan, Function, Not
    from ._basis import LinearBasisFunction, HingeBasisFunction, SmoothedHingeBasisFunction, \
          MissingnessBasisFunction, ConstantBasisFunction, VariableBasisFunction

    Missing = Function('Missing')
    NaNProtect = Function('NaNProtect')

    def linear_bf_to_factor(bf, bf_var):
        return bf_var

    def smoothed_hinge_bf_to_factor(bf, bf_var):
        knot = RealNumber(bf.get_knot())
        knot_minus = RealNumber(bf.get_knot_minus())
        knot_plus = RealNumber(bf.get_knot_plus())
        r = RealNumber(bf.get_r())
        p = RealNumber(bf.get_p())
        if bf.get_reverse():
            lower_p = (-(bf_var - knot)), (bf_var <= knot_minus)
            upper_p = (0, bf_var >= knot_plus)
            left_exp = Mul(p, Pow((bf_var - knot_plus), 2))
            right_exp = Mul(r, Pow((bf_var - knot_plus), 3))
            middle_b = And(knot_minus < bf_var, bf_var < knot_plus)
            middle_exp = (Add(left_exp, right_exp), middle_b)
            piecewise = Piecewise(lower_p, upper_p, middle_exp)
            factor = piecewise
        else:
            lower_p = (0, bf_var <= knot_minus)
            upper_p = (bf_var - knot, bf_var >= knot_plus)
            left_exp = Mul(p, Pow((bf_var - knot_minus), 2))
            right_exp = Mul(r, Pow((bf_var - knot_minus), 3))
            middle_b = And(knot_minus < bf_var, bf_var < knot_plus)
            middle_exp = (Add(left_exp, right_exp), middle_b)
            piecewise = Piecewise(lower_p, upper_p, middle_exp)
            factor = piecewise
        return factor

    def hinge_bf_to_factor(bf, bf_var):
        knot = bf.get_knot()
        if bf.get_reverse():
            factor = Max(0, RealNumber(knot) - bf_var)
        else:
            factor = Max(0, bf_var - RealNumber(knot))
        return factor

    def missingness_bf_to_factor(bf, bf_var):
        # This is the error that should be raised when a user attempts to use functionality
        # that has not yet been implemented.
        if bf.complement:
            return Not(Missing(bf_var))
        else:
            return Missing(bf_var)

    def constant_bf_to_factor(bf, bf_var):
        return RealNumber(1)

    def protect_from_nan(label, missables):
        return NaNProtect(Symbol(label)) if label in missables else Symbol(label)

    def dont_protect_from_nan(label, missables):
        return Symbol(label)

    bf_to_factor_dispatcher = {LinearBasisFunction: linear_bf_to_factor,
                               SmoothedHingeBasisFunction: smoothed_hinge_bf_to_factor,
                               HingeBasisFunction: hinge_bf_to_factor,
                               MissingnessBasisFunction: missingness_bf_to_factor,
                               ConstantBasisFunction: constant_bf_to_factor}

    nan_protect_dispatch = {LinearBasisFunction: protect_from_nan,
                            SmoothedHingeBasisFunction: protect_from_nan,
                            HingeBasisFunction: protect_from_nan,
                            MissingnessBasisFunction: dont_protect_from_nan,
                            ConstantBasisFunction: protect_from_nan}

    def bf_to_factor(bf, missables):
        '''
        Convert a BasisFunction to a factor of a term.
        '''
        if isinstance(bf, VariableBasisFunction):
            bf_var = nan_protect_dispatch[bf.__class__](bf.label, missables)

        else:
            bf_var = None
        return bf_to_factor_dispatcher[bf.__class__](bf, bf_var)

    def missingness_bf_get_missables(bf):
        bf_var = bf.label
        return set([bf_var])

    def non_missable(bf):
        return set()

    bf_get_missables_dispatcher = {LinearBasisFunction: non_missable,
                                   SmoothedHingeBasisFunction: non_missable,
                                   HingeBasisFunction: non_missable,
                                   MissingnessBasisFunction: missingness_bf_get_missables,
                                   ConstantBasisFunction: non_missable}

    def get_missables(bf):
        missables = bf_get_missables_dispatcher[bf.__class__](bf)
        parent = bf.get_parent()
        if parent is None:
            return missables
        else:
            missables.update(get_missables(parent))

        return missables

    def bf_to_term(bf, missables):
        '''
        Convert a BasisFunction to a term (without coefficient).
        '''
        term = bf_to_factor(bf, missables)
        parent = bf.get_parent()
        if parent is None:
            return term
        else:
            return Mul(term, bf_to_term(parent, missables))

    return [bf_to_term(bf, get_missables(bf)) for bf in earth_model.basis_.piter()]


def export_sympy(earth_model, columns=None):
    """
    Constructs a sympy expression or list of sympy expressions from of a trained earth model.

    :param earth_model: Trained pyearth model
    :param columns: The index or indices of the output columns for which expressions are to
      be constructed.  If an integer is used, a sympy expression is returned.  If indices
      are given then a list of sympy expressions is returned.  If columns is None, it is treated
      as if columns=0 for models with only one output column or as columns=slice(None) for more than
      one output column.
    :return: a sympy expression or list of sympy expressions equivalent to the Earth.predict method for
      the selected output columns.

    """
    # Set a sane default for columns
    if columns is None:
        if earth_model.coef_.shape[0] == 1:
            columns = 0
        else:
            columns = slice(None)

    # Get basis function terms
    terms = export_sympy_term_expressions(earth_model)

    # Handle column choice
    coefs = earth_model.coef_[columns]
    if len(coefs.shape) == 1:
        unwrap = True
        coefs = [coefs]
        n_cols = 1
    else:
        unwrap = False
        n_cols = coefs.shape[0]

    # Combine coefficients with terms for each output column
    result = [sum([coefs[i][j] * term for j, term in enumerate(terms)]) for i in range(n_cols)]

    if unwrap:
        # Result should be an expression rather than a list of expressions.
        result = result[0]
    return result







import os
from functools import wraps
from nose import SkipTest
from nose.tools import assert_almost_equal
from distutils.version import LooseVersion
import sys

def if_environ_has(var_name):
    # Test decorator that skips test if environment variable is not defined
    def if_environ(func):
        @wraps(func)
        def run_test(*args, **kwargs):
            if var_name in os.environ:
                return func(*args, **kwargs)
            else:
                raise SkipTest('Only run if %s environment variable is '
                               'defined.' % var_name)
        return run_test
    return if_environ

def if_platform_not_win_32(func):
    @wraps(func)
    def run_test(*args, **kwargs):
        if sys.platform == 'win32':
            raise SkipTest('Skip for 32 bit Windows platforms.')
        else:
            return func(*args, **kwargs)
    return run_test
            
def if_sklearn_version_greater_than_or_equal_to(min_version):
    '''
    Test decorator that skips test unless sklearn version is greater than or
    equal to min_version.
    '''
    def _if_sklearn_version(func):
        @wraps(func)
        def run_test(*args, **kwargs):
            import sklearn
            if LooseVersion(sklearn.__version__) < LooseVersion(min_version):
                raise SkipTest('sklearn version less than %s' %
                               str(min_version))
            else:
                return func(*args, **kwargs)
        return run_test
    return _if_sklearn_version


def if_statsmodels(func):
    """Test decorator that skips test if statsmodels not installed. """

    @wraps(func)
    def run_test(*args, **kwargs):
        try:
            import statsmodels
        except ImportError:
            raise SkipTest('statsmodels not available.')
        else:
            return func(*args, **kwargs)
    return run_test


def if_pandas(func):
    """Test decorator that skips test if pandas not installed. """

    @wraps(func)
    def run_test(*args, **kwargs):
        try:
            import pandas
        except ImportError:
            raise SkipTest('pandas not available.')
        else:
            return func(*args, **kwargs)
    return run_test

def if_sympy(func):
    """ Test decorator that skips test if sympy not installed """ 
    
    @wraps(func)
    def run_test(*args, **kwargs):
        try:
            from sympy import Symbol, Add, Mul, Max, RealNumber, Piecewise, sympify, Pow, And, lambdify
        except ImportError:
            raise SkipTest('sympy not available.')
        else:
            return func(*args, **kwargs)
    return run_test
    


def if_patsy(func):
    """Test decorator that skips test if patsy not installed. """

    @wraps(func)
    def run_test(*args, **kwargs):
        try:
            import patsy
        except ImportError:
            raise SkipTest('patsy not available.')
        else:
            return func(*args, **kwargs)
    return run_test


def assert_list_almost_equal(list1, list2):
    for el1, el2 in zip(list1, list2):
        assert_almost_equal(el1, el2)


def assert_list_almost_equal_value(list, value):
    for el in list:
        assert_almost_equal(el, value)
